
(module dw mzscheme
  (require "redex.scm"
           (lib "pretty.ss")
           (lib "list.ss")
           (lib "match.ss")
           "trim.scm")

  (provide reductions
           lang)
  
  (define lang
    (language (p (store (sf ...) (dw (dws ...) e)))
              (sf (x v))
              (dws (x e e))
              
              (e (e e ...) (set! x e) (begin e e ...) (if e e e) x v
                 (push dws)
                 (pop)
                 (throw dws ... e))
              
              (v (lambda (x ...) e) + * number unspecified #t #f
                 call/cc dynamic-wind)
              
              (P (store (sf ...) (dw (dws ...) E)))
              (E (v ... E e ...)
                 (begin E e e ...)
                 (set! x E)
                 (if E e e)
                 hole)
              
              (x (variable-except set! begin if lambda + * push pop call/cc dynamic-wind throw unspecified))))
  
  (define reductions
    (reduction-relation
     lang
     (--> (in-hole P_1
                   (dynamic-wind (lambda () e_1)
                                 (lambda () e_2)
                                 (lambda () e_3)))
          (in-hole P_1
                   (begin e_1
                          (push (x_1 e_1 e_3))
                          ((lambda (x_2) (begin (pop) e_3 x_2))
                           e_2)))
          "DWWind"
          (fresh x_1 x_2))
     
     (--> (store (sf_1 ...) (dw (dws_1 ...) (in-hole E_1 (push dws_2))))
          (store (sf_1 ...) (dw (dws_1 ... dws_2) (in-hole E_1 unspecified)))
          "DWPush")
     
     (--> (store ((x_s v_s) ...) (dw (dws_1 ... dws_n) (in-hole E_1 (pop))))
          (store ((x_s v_s) ...) (dw (dws_1 ...) (in-hole E_1 unspecified)))
          "DWPop")
     
     ;; call/cc
     (--> (store (sf_1 ...) (dw (dws_1 ...) (in-hole E_1 (call/cc v_1))))
          (store (sf_1 ...)
            (dw (dws_1 ...) 
                (in-hole E_1
                         (v_1 
                          (lambda (x)
                            (throw dws_1 ... (in-hole E_1 x)))))))
          "DWCallcc"
          (fresh x))
     
     ;; throw to a continuation
     (--> (store ((x_s v_s) ...) (dw (dws_1 ...) (in-hole E_1 (throw dws_2 ... e_1))))
          (store ((x_s v_s) ...) (dw (dws_1 ...) (begin ,@(trim (term (dws_1 ...)) (term (dws_2 ...)))
                                                        e_1)))
          "DWThrow")
     
     ;; generic reductions
     (~~> ((lambda (x_1 ...) e_1) v_1 ...) 
          ,(subst-all (term (x_1 ...)) (term (v_1 ...)) (term (begin e_1)))
          "beta-v")
     
     (~~> (+ number_1 ...) ,(apply + (term (number_1 ...))) "+")
     (~~> (* number_1 ...) ,(apply * (term (number_1 ...))) "*")
     
     (~~> (begin v e_1 e_2 ...) (begin e_1 e_2 ...) "beginc")
     (~~> (begin e_1) e_1 "begind")

     (~~> (if #f e_1 e_2) e_2 "ift")
     (~~> (if v_1 e_1 e_2)
          e_1
          (side-condition (not (eq? (term v_1) #f)))
          "iff")
     
     (--> (store ((x_b v_b) ... (x_1 v_1) (x_a v_a) ...)
                 (dw (dws_1 ...)
                     (in-hole E_1 (set! x_1 v_2))))
          (store ((x_b v_b) ... (x_1 v_2) (x_a v_a) ...)
                 (dw (dws_1 ...)
                     (in-hole E_1 unspecified)))
          "set!")
     (--> (store ((x_b v_b) ... (x_1 v_1) (x_a v_a) ...)
                 (dw (dws_1 ...)
                     (in-hole E_1 x_1)))
          (store ((x_b v_b) ... (x_1 v_1) (x_a v_a) ...)
                 (dw (dws_1 ...)
                     (in-hole E_1 v_1)))
          "varref")
     

     where
     [(~~> a b) (--> (in-hole P_1 a) (in-hole P_1 b))]))

  (define subst-one 
    (subst
     [`throw (constant)]
     [`call/cc (constant)]
     [(? symbol?) (variable)]
     [(? number?) (constant)]
     [(? boolean?) (constant)]
     [`(lambda ,xs ,@(bodies ...))
       (all-vars xs)
       (build (lambda (vars . bodies) `(lambda ,vars ,@bodies)))
       (subterms xs bodies)]
     [`(set! ,x ,e)
       (all-vars '())
       (build (lambda (vars x exp) `(set! ,x ,exp)))
       (subterm '() x)
       (subterm '() e)]
     [(f args ...)
      (all-vars '())
      (build (lambda (vars f . args) `(,f ,@args)))
      (subterm '() f)
      (subterms '() args)]))
  
  (define (subst-all params args body) (foldl subst-one body params args)))
