(module generic-fixed-repl mzscheme
  
  (require "tl-tests.scm"
           "test.scm"
           (lib "port.ss"))
  
  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  ;;
  ;;  code for the paper
  ;; (indented to the edge for ease of copy/paste)
  ;;
  
(define (b-eval expr)
  (if (and (pair? expr) 
           (eq? (car expr)
                'begin))
      (evals (cdr expr))
      (eval expr)))
  
(define (evals exprs)
  (let ((answers 'dummy))
    (let loop ()
      (if (pair? exprs)
          (let ((expr (car exprs)))
            (set! exprs (cdr exprs))
            (call-with-values
             (lambda () (b-eval expr))
             (lambda args (set! answers args)))
            (loop))
          (apply values answers)))))

#;  
(define (evals exprs)
  (let loop ()
    (if (pair? exprs)
        (let ((expr (car exprs)))
          (set! exprs (cdr exprs))
          (b-eval expr)
          (loop)))))
  
  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  ;;
  ;;  testing the evaluator
  ;;
  
  (define (repl)
    (let ([final-answer #f])
      (let loop ()
        (let ([expr (read)])
          (if (eof-object? expr)
              final-answer
              (begin
                (set! final-answer (b-eval expr))
                (loop)))))))
  
  (define (test-repl/proc exprs vars)
    (parameterize ([current-namespace (make-namespace)])
      (let-values ([(in out) (make-pipe)])
        (thread
         (λ () 
           (for-each (λ (x) (write x out) (newline out))
                     exprs)
           (close-output-port out)))
        (let ([answer 
               (with-handlers ([exn? (λ (x) (massage-exn-message (exn-message x)))])
                 (parameterize ([current-input-port in])
                   (repl)))])
          (values (map namespace-variable-value vars)
                  answer)))))
  
  (define (massage-exn-message x) 
    (regexp-replace 
     #rx"undefined"
     (regexp-replace 
      #rx"set!: cannot set undefined identifier"
      x
      "attempt to set! free identifier")
     "free"))
  
  (define test-count 0)
  (define (run-test test)
    (let* ([in (test-input test)]
           [out (if (= 1 (length (test-expecteds test)))
                    (car (test-expecteds test))
                    (error 'generic-fixed-repl.scm "expected a single possible result, got ~s" (test-expecteds test)))]
           [store-vars (map car (cadr out))]
           [store-expecteds (map cadr (cadr out))]
           [final-expected (caddr out)]
           [body-exprs (cddr in)])
      (let-values ([(store-gots final-got) (test-repl/proc (rewrite-test body-exprs) store-vars)])
        (set! test-count (+ test-count 1))
        (unless (andmap compare-results? store-expecteds store-gots)
          (printf "~s (~a:~a) failed\n" 
                  (test-name test)
                  (test-file test)
                  (test-line test))
          (printf "store     vars ~s\n" store-vars)
          (printf "store      got ~s\n" store-gots)
          (printf "store expected ~s\n" store-expecteds)
          (printf "\n"))
        (unless (compare-results? final-expected final-got)
          (printf "~s (~a:~a) failed\n" 
                  (test-name test)
                  (test-file test)
                  (test-line test))
          (printf "     got ~s\n" final-got)
          (printf "expected ~s\n" final-expected)
          (printf "\n")))))
  
  (define (compare-results? x y)
    (or (procedure? y)
        (and (eq? x 'unspecified)
             (equal? y (void)))
        (equal? x y)))
  
  (define (rewrite-test exp)
    (cond
      [(pair? exp) (cons (rewrite-test (car exp))
                         (rewrite-test (cdr exp)))]
      [(eq? exp 'beginD) 'begin]
      [(eq? exp 'beginE) 'begin]
      [else exp]))
  
  (for-each run-test (test-suite-tests tl-ts))
  (printf "ran ~a tests\n" test-count))
